/* NasTalk - Communicate with a Terminal Server using the Telnet protocol.
 * Copyright (C) 2000, Robert Davidson.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* Compile with:
 * g++ acua_nastalk.cc -o acua_nastalk
 */

#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#ifndef SIGHUP
#include <signum.h>
#endif

// -- User defines ---------------------------------------------------

// Path to netcat (a hacked up version comes with RADACUA).
#define NETCAT          "/root/radacua/netcat/nc"
// Path to the "expect" program.  "chat" will not work without modifications.
#define EXPECT          "/usr/bin/expect"
// Maximum time to wait for both processes to return (secs*2)
#define TIMEOUT         20   // 20 = 10 seconds.
// DEBUG mode (1=on, 0=off)
#define DEBUG           1

//-- End user defined stuff ------------------------------------------


int main (int argc, char **argv)
{
  int                   cstat=0, cld_ret=0, cld_timeout=0, cp=0, pipes_a[2], pipes_b[2];
  pid_t                 ncpid, cpid;
  struct termios        ti_a, ti_b;

  // check that there is the right amount of args.
  if (argc != 3) {
    printf("incorrect usage of acua_nastalk.\n");
    printf("usage: acua_nastalk <hostname> <path_to_script>\n");
    return(1);
  }

  if (pipe(pipes_a)) {
    // error making pipes
    syslog(LOG_ERR, "Could not create pipe.");
    return(1);
  }

  if (pipe(pipes_b)) {
    // error making pipes
    syslog(LOG_ERR, "Could not create pipe.");
    return(1);
  }

  switch (ncpid = fork()) {
  case -1:
    // fork failed - parent process
    syslog(LOG_ERR, "Could not fork.");
    return(-1);
  case 0:
    // we're the child process
    close(1);             // need to close this
    dup(pipes_a[1]);      // make stdout the same as pipes_a[1]
    close(pipes_a[0]);    // dont need this one
    // ----
    close(0);
    dup(pipes_b[0]);
    close(pipes_b[1]);
    // launch netcat.
    if (DEBUG) syslog(LOG_INFO, "starting netcat");
    execl(NETCAT, "nc", "-t", "-w", "20", argv[1], "23", (char*)NULL);
    exit(0);
  default:
    // parent process.
    // lets fork another damn process.
    switch (cpid = fork()) {
    case -1:
      // fork failed
      syslog(LOG_ERR, "c2: fork failed.");
      return(1);
    case 0:
      // we're the second child process
      close(0);             // not needed
      dup(pipes_a[0]);      // make stdin the same as pipes_a[0]
      close(pipes_a[1]);    // not needed
      // ----
      close(1);
      dup(pipes_b[1]);
      close(pipes_b[0]);
      sleep(1);
      if (DEBUG) syslog(LOG_INFO, "starting expect");
      execl(EXPECT, "expect", argv[2], (char*)NULL);
      exit(0);
    default:
      // parent process
      // make sure we:  1) Never crash.
      //                2) Fix crashed children with a kill.
      //                3) Try to clean up after the little brats.
      while (cp < 2) {
        cld_timeout++;
        cld_ret = waitpid(-1, &cstat, WNOHANG);
        // debugging stuff, just incase something plays games.
        if (DEBUG) {
           if (cld_ret == cpid) syslog(LOG_INFO, "expect process returned. err = %d", WIFEXITED(cstat));
           if (cld_ret == ncpid) syslog(LOG_INFO, "netcat process returned. err = %d", WIFEXITED(cstat));
        }
        if ((cld_ret == -1) || (cld_ret > 0)) cp++;
        if (cp < 2) usleep(500000);  // half a second pause
        if ((cld_timeout == TIMEOUT) || (cld_ret == 2)) {
          if (cld_timeout == TIMEOUT) {
            if (DEBUG)
              syslog(LOG_ERR, "timeout waiting for child processes to finish.");
            // clean up the mess.
            kill(cpid, SIGKILL);       // bugger off.
            kill(ncpid, SIGKILL);
            usleep(500000);  // half a second pause
            // not sure if we need to do this, or infact should.
            // the aim is to clean up any zombie processes we've left around.
            for (cp=0;cp<=10;cp++) waitpid(-1, NULL, WNOHANG);
            return(1);
          } else if (DEBUG) syslog(LOG_INFO, "exiting normally.");
        }
      }
    }
  }
  return(0);
}

